{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "3f168073-83b5-445f-80e4-d0f1ff3d4cb9",
   "metadata": {},
   "source": [
    "# Unit Testing : Making sure your bugs don't come back\n",
    "\n",
    "Unit testing allows to ensure that a given software behaves in the\n",
    "correct way, at least for the cases one is testing. Once a function is\n",
    "written (or even before in TTD) or a bug is fixed, it is necessary to\n",
    "write a test that ensures the function to work properly in limit cases\n",
    "or the bug to not reappear in the future. There are several levels\n",
    "associated with [unit testing\n",
    ".](https://en.wikipedia.org/wiki/Unit_testing)\n",
    "\n",
    "In this unit we will learn the general philosophy behind it and a couple\n",
    "of tools to implement very basic tests, althoguh the list of [testing\n",
    "frameworks](https://en.wikipedia.org/wiki/List_of_unit_testing_frameworks)\n",
    "is very large. Furthermore, modularization will be very important, so\n",
    "you must have a clear understanding on how to split some given code into\n",
    "headers, source files, and how to compile objects and then link them\n",
    "using the linker, hopefully through a Makefile.\n",
    "\n",
    "It is worth mentioning that catching an exception (`try` and `catch` blocks) also helps with handling runtime errors and reacting correspondingly. Also, using logging libraries like logger or loguru (python), or `spdlog`, https://github.com/gabime/spdlog, (c++) allows to print useful logging messages that also helps when you try to understand what is going on with your program.  \n",
    "\n",
    "<https://matklad.github.io/2021/05/31/how-to-test.html>"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "f1ec0607",
   "metadata": {},
   "source": [
    "\n",
    "## Catch2\n",
    "\n",
    "Our goal here is to learn to use\n",
    "[catch2](https://github.com/catchorg/Catch2/blob/master/docs/tutorial.md#top)\n",
    "to test a very simple function extracted from their tutorial. Later we\n",
    "will modularize the code to practice that and write a useful `Makefile`. \n",
    "\n",
    "### Install catch2\n",
    "\n",
    "If you don't have it installed, you can use `spack`:\n",
    "\n",
    "``` shell\n",
    "spack install catch2\n",
    "```\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "c1c4947d",
   "metadata": {},
   "source": [
    "\n",
    "### Tutorial example: factorial\n",
    "\n",
    "Here we will follow the tutorial , testing a factorial function implementation. To do so, we need, first, to modularize our code:\n",
    "- `factorial.h` :"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "2e5101e7",
   "metadata": {},
   "source": [
    "\n",
    "  ```c++\n",
    "  #pragma once\n",
    "  int factorial(int n);\n",
    "  ```\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "0d2ec098",
   "metadata": {},
   "source": [
    "\n",
    "- `factorial.cpp`:\n",
    "\n",
    "  ```c++\n",
    "    #include \"factorial.h\"\n",
    "\n",
    "    int factorial(int number)\n",
    "    {\n",
    "        return number <= 1 ? number : factorial(number-1)*number;\n",
    "    }\n",
    "  ```  \n",
    "\n",
    "- `main_factorial.cpp`\n",
    "  ```c++\n",
    "    #include <iostream>\n",
    "    #include \"factorial.h\"\n",
    "\n",
    "    int main(void)\n",
    "    {\n",
    "    std::cout << factorial(4) << std::endl;\n",
    "    return 0;\n",
    "  }\n",
    "  ```\n",
    "With these three files we have the basic utilities to use the factorial function. To compile, we must run something like\n",
    "``` shell\n",
    "g++ -c factorial.cpp\n",
    "g++ -c main_factorial.cpp\n",
    "g++ factorial.o main_factorial.o -o factorial_test.x\n",
    "```\n",
    "and then run as\n",
    "```shell\n",
    "./factorial_test.x\n",
    "```\n",
    "The compilation can be automated with a  `Makefile` as\n",
    "\n",
    "```Makefile\n",
    "all: main_factorial.x\n",
    "\n",
    "%.x: %.o factorial.o\n",
    "\tg++ $^ -o $@\n",
    "\n",
    "%.o: %.cpp\n",
    "\tg++ -c $<\n",
    "\n",
    "clean:\n",
    "\trm -f *.o *.x\n",
    "```\n",
    "\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "84329ad4",
   "metadata": {},
   "source": [
    "### Including a test using catch2\n",
    "\n",
    "This is the file example extracted from `catch2` tutorial.\n",
    "\n",
    "``` cpp\n",
    "#define CATCH_CONFIG_MAIN  // This tells Catch to provide a main() - only do this in one cpp file\n",
    "#include \"catch2/catch_test_macros.hpp\"\n",
    "\n",
    "#include \"factorial.h\"\n",
    "\n",
    "TEST_CASE( \"Factorials are computed\", \"[factorial]\" ) {\n",
    "    //REQUIRE( factorial(0) == 1 );\n",
    "    REQUIRE( factorial(1) == 1 );\n",
    "    REQUIRE( factorial(2) == 2 );\n",
    "    REQUIRE( factorial(3) == 6 );\n",
    "    REQUIRE( factorial(10) == 3628800 );\n",
    "}\n",
    "```\n",
    "To compile, you need to also link with the corresponging catch2 flags (if you are using spack, do not forget to load catch2, `spack load catch2`)\n",
    "\n",
    "```bash\n",
    "g++ -c test_factorial.cpp\n",
    "g++ -c factorial.cpp\n",
    "g++ test_factorial.o factorial.o -o test_factorial.x -l Catch2Main -l Catch2\n",
    "./test_factorial.x\n",
    "```\n",
    "The last two flags, `-l Catch2Main -l Catch2`, are used to link the program with the catch implementations. Sometimes, for non standard installations, it is useful to configure the paths to find both the includes and libs, and this can be done with the `pkg-config` utility. For example, to get the include path one can use\n",
    "```shell\n",
    "$ pkg-config --cflags catch2\n",
    "-I/usr/local/include # This result can change if catch2 is installed on other systems\n",
    "```\n",
    "or for the libs path\n",
    "```shell\n",
    "$ pkg-config --libs-only-L catch2\n",
    "-L/usr/local/lib  # This result can change if catch2 is installed on other systems\n",
    "```\n",
    "So, the compilation line could \n",
    "```bash\n",
    "g++ $(pkg-config --cflags) $(pkg-config --libs-only-L catch2) test_factorial.o factorial.o -o test_factorial.x -l Catch2Main -l Catch2\n",
    "```\n",
    "Again, it is better to include this in a Makefile:\n",
    "```Makefile\n",
    "SHELL:=/bin/bash\n",
    "\n",
    "all: main_factorial.x\n",
    "\n",
    "%.x: %.o factorial.o\n",
    "\tg++ $^ -o $@\n",
    "\n",
    "test_factorial.x: test_factorial.o factorial.o\n",
    "\tg++ $$(pkg-config --cflags) $$(pkg-config --libs-only-L catch2) $^ -o $@ -l Catch2Main -l Catch2\n",
    "\n",
    "test: test_factorial.x\n",
    "\t./test_factorial.x\n",
    "\n",
    "%.o: %.cpp\n",
    "\tg++ -c $<\n",
    "\n",
    "clean:\n",
    "\trm -f *.o *.x\n",
    "```\n",
    "\n",
    "This way, when you write \n",
    "```bash\n",
    "make test\n",
    "```\n",
    "it will compile and then run all the test. \n",
    "\n",
    "Please uncomment the commented line in the test, analyze and fix the factorial function. Also implemente more tests, for large numbers, negative numbers, and so on.\n",
    "\n",
    "NOTE: if you are using spack, you might need to add more code to your target commands\n",
    "```Makefile\n",
    "%.x: %.o factorial.o\n",
    "    source $$HOME/repos/spack/share/spack/setup-env.sh; \\\n",
    "    spack load catch2; \\\n",
    "    g++ $$(pkg-config --cflags catch2) $^ -o $@\n",
    "```\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "8cd4d954",
   "metadata": {},
   "source": [
    "\n",
    "## Google test\n",
    "\n",
    "Google test is a famous and advance unit framework that goes well beyond\n",
    "of what is shown here. You are invited to follow the\n",
    "[docs](https://github.com/google/googletest/blob/master/googletest/docs/primer.md)\n",
    "to learn more.\n",
    "\n",
    "### Installation\n",
    "\n",
    "Again, we will use spack\n",
    "\n",
    "``` shell\n",
    "spack install googletest\n",
    "mkdir googletest\n",
    "```\n",
    "\n",
    "### Example\n",
    "\n",
    "This is an example, already modularized.\n",
    "\n",
    "-   Factorial and isprime header:\n",
    "\n",
    "``` cpp\n",
    "#ifndef GTEST_SAMPLES_SAMPLE1_H_\n",
    "#define GTEST_SAMPLES_SAMPLE1_H_\n",
    "\n",
    "// Returns n! (the factorial of n).  For negative n, n! is defined to be 1.\n",
    "int Factorial(int n);\n",
    "\n",
    "//// Returns true if and only if n is a prime number.\n",
    "bool IsPrime(int n);\n",
    "\n",
    "#endif  // GTEST_SAMPLES_SAMPLE1_H_\n",
    "```\n",
    "\n",
    "-   Source file\n",
    "\n",
    "``` cpp\n",
    "#include \"factorial.h\"\n",
    "\n",
    "// Returns n! (the factorial of n).  For negative n, n! is defined to be 1.\n",
    "    int Factorial(int n) {\n",
    "    int result = 1;\n",
    "    for (int i = 1; i <= n; i++) {\n",
    "        result *= i;\n",
    "    }\n",
    "\n",
    "    return result;\n",
    "  }\n",
    "\n",
    "// Returns true if and only if n is a prime number.\n",
    "bool IsPrime(int n) {\n",
    "    // Trivial case 1: small numbers\n",
    "    if (n <= 1) return false;\n",
    "\n",
    "    // Trivial case 2: even numbers\n",
    "    if (n % 2 == 0) return n == 2;\n",
    "\n",
    "    // Now, we have that n is odd and n >= 3.\n",
    "\n",
    "    // Try to divide n by every odd number i, starting from 3\n",
    "    for (int i = 3; ; i += 2) {\n",
    "        // We only have to try i up to the square root of n\n",
    "        if (i > n/i) break;\n",
    "\n",
    "        // Now, we have i <= n/i < n.\n",
    "        // If n is divisible by i, n is not prime.\n",
    "        if (n % i == 0) return false;\n",
    "    }\n",
    "\n",
    "    // n has no integer factor in the range (1, n), and thus is prime.\n",
    "    return true;\n",
    "}\n",
    "```\n",
    "\n",
    "-   Test source file (to be compiled as an object)\n",
    "\n",
    "``` cpp\n",
    "#include <limits.h>\n",
    "#include \"factorial.h\"\n",
    "#include \"gtest/gtest.h\"\n",
    "namespace {\n",
    "    // Tests factorial of negative numbers.\n",
    "    TEST(FactorialTest, Negative) {\n",
    "        // This test is named \"Negative\", and belongs to the \"FactorialTest\"\n",
    "        // test case.\n",
    "        EXPECT_EQ(1, Factorial(-5));\n",
    "        EXPECT_EQ(1, Factorial(-1));\n",
    "        EXPECT_GT(Factorial(-10), 0);\n",
    "    }\n",
    "    // Tests factorial of 0.\n",
    "    TEST(FactorialTest, Zero) {\n",
    "        EXPECT_EQ(1, Factorial(0));\n",
    "    }\n",
    "\n",
    "// Tests factorial of positive numbers.\n",
    "    TEST(FactorialTest, Positive) {\n",
    "        EXPECT_EQ(1, Factorial(1));\n",
    "        EXPECT_EQ(2, Factorial(2));\n",
    "        EXPECT_EQ(6, Factorial(3));\n",
    "        EXPECT_EQ(40320, Factorial(8));\n",
    "    }\n",
    "\n",
    "    // Tests negative input.\n",
    "    TEST(IsPrimeTest, Negative) {\n",
    "        // This test belongs to the IsPrimeTest test case.\n",
    "\n",
    "        EXPECT_FALSE(IsPrime(-1));\n",
    "        EXPECT_FALSE(IsPrime(-2));\n",
    "        EXPECT_FALSE(IsPrime(INT_MIN));\n",
    "    }\n",
    "\n",
    "// Tests some trivial cases.\n",
    "    TEST(IsPrimeTest, Trivial) {\n",
    "        EXPECT_FALSE(IsPrime(0));\n",
    "        EXPECT_FALSE(IsPrime(1));\n",
    "        EXPECT_TRUE(IsPrime(2));\n",
    "        EXPECT_TRUE(IsPrime(3));\n",
    "    }\n",
    "\n",
    "// Tests positive input.\n",
    "    TEST(IsPrimeTest, Positive) {\n",
    "        EXPECT_FALSE(IsPrime(4));\n",
    "        EXPECT_TRUE(IsPrime(5));\n",
    "        EXPECT_FALSE(IsPrime(6));\n",
    "        EXPECT_TRUE(IsPrime(23));\n",
    "    }\n",
    "}\n",
    "```\n",
    "\n",
    "-   Main google test file\n",
    "\n",
    "``` cpp\n",
    "#include <cstdio>\n",
    "#include \"gtest/gtest.h\"\n",
    "\n",
    "GTEST_API_ int main(int argc, char **argv) {\n",
    "    printf(\"Running main() from %s\\n\", __FILE__);\n",
    "    testing::InitGoogleTest(&argc, argv);\n",
    "    return RUN_ALL_TESTS();\n",
    "}\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "b7475bfe",
   "metadata": {},
   "source": []
  }
 ],
 "metadata": {
  "language_info": {
   "name": "python"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
